#ProcessorEnvironment.py
'''Functions to process the environment component.'''

import logging
import os
from distutils.dir_util import copy_tree
from glob import glob
from pathlib import Path

from coordinator import EnvSuitPipeline
from coordinator.ProcessorServer import (
    get_data_from_server,
)
from coordinator.ProcessorUtils import add_filters_to_sublogger, short_name, get_only_existing_globs
from coordinator.processor_base import ProcessorBase
from ews_postprocessing.environmental_suitability.env_suit_post_processor import EnvSuitPostProcessor
from ews_postprocessing.utils.disease_info import EnvSuitDiseaseInfo


logger = logging.getLogger(__name__)

class ProcessorEnvironment(ProcessorBase):

    def set_component_logger(self):
        logger = logging.getLogger('Processor.Environment')
        add_filters_to_sublogger(logger)

    def process_pre_job(self, args):
        # return process_pre_job_server_download(args)
        return True

    def process_in_job(self, jobPath, status, configjson, component) -> object:
        return self.process_in_job_env2_0(jobPath, status, configjson, component)


    def process_post_job(self, jobPath, configjson) -> [str]:
        return self.process_EWS_plotting_env2_0(jobPath, configjson)


    def __init__(self) -> None:
        super().__init__()

    def process_in_job_env2_0(self, jobPath,status,config,component) -> dict:
        '''Download met data from remote, prepare data, and run :class:`EnvSuitPipeline` pipeline.'''
    
        logger.info('started process_in_job_env2_0()')
    
        # Initialise output variable
        proc_out = {}
        # Output files available for upload    
        proc_out['output'] = None
        # Processing files available for clearing
        proc_out['clearup'] = []

        logger.info('Copying file from remote server to job directory')
    
        data_result = get_data_from_server(jobPath,config,component)

        if data_result['output'] is not None:
            proc_out['output'] = data_result['output']

        proc_out['clearup'] += data_result['clearup']
    
        region = config['RegionName']
    
        logger.info(f"Calling environmental suitability 2.0 for {region} so wait for output to appear")
    
        pipeline_config = config["Environment"]
        try:
            #todo lawrence comment this back to original (extracted=False)
            EnvSuitPipeline.run_pipeline(pipeline_config, region, config["StartString"], extracted=False)
        except:
            logger.exception(f"Some failure when running EnvSuitPipeline.py")
            raise
    
        logger.info('Finished running environmental suitability 2.0')
    
        # TODO: Check that the output appears as expected

        return proc_out
    
    def process_copy_past_job_env2_0(self, jobPath,status,config,component):
        '''For when we want to skip process_in_job() to test the other components of
        this script. Currently hard-wired.'''
    
        # TODO: remove this hard-wired assumption
        jobPath_to_copy = f"{jobPath}/../{short_name['Environment']}_{config['StartString']}_bak/"
    
        assert os.path.exists(jobPath_to_copy)
    
        dir_src = f"{jobPath_to_copy}/processed/"
    
        dir_dst = f"{jobPath}/processed/"
    
        logger.info(f"Copying from {dir_src}")
    
        logger.info(f"to {dir_dst}")
    
        copy_tree(dir_src,dir_dst)
    
        logger.info('Copying complete')
    
        proc_out = {}
        # Output files available for upload
        proc_out['output'] = None
        # Processing files available for clearing
        proc_out['clearup'] = None
    
        return proc_out
    
    '''class EWSPlottingEnvSuit(EWSPlottingEnvSuitBase):
    
        def set_custom_params(self,
                              sys_params_dict: dict,
                              chart_params_dict: dict,
                              run_params_dict: dict,
                              disease_csv_template_arg: str,
                              diseases: List[EnvSuitDiseaseInfo]):
            # this is unique to the asia/east africa env suit, as we are not filtering within country boundaries
            run_params_dict[RUN_PARAMS.FILTER_FOR_COUNTRY_KEY] = "False"'''
    
    #TODO test if this works
    def process_EWS_plotting_env2_0(self, jobPath,config) -> [str]:
        '''Configures the plotting arguments and calls EWS-plotting as a python module.
        Returns a list of output files for transfer.'''
    
        logger.info('started process_EWS_plotting_env2_0()')
    
        main_region = config['RegionName']
    
        input_dir = f"{jobPath}/processed/{main_region}"
    
        subregions = config['SubRegionNames']
    
        ews_plotting_output_globs = []
    
        # work on each region
        for region in subregions:
    
            output_dir = f"{jobPath}/plotting/{region.lower()}"
            rie_template_dir = input_dir + "/{DISEASE_DIR}/RIE_value.nc"

            Path(output_dir).mkdir(parents=True, exist_ok=True)
    
            sys_config = config['Environment']['EWS-Plotting']['SysConfig']
            run_config = config['Environment']['EWS-Plotting']['RunConfig']
            chart_config = config['Environment']['EWS-Plotting'][region]['ChartConfig']
            filter_for_country = config['Environment']['EWS-Plotting'][region]['FilterForCountry']
    
            # Note that this runs all disease types available
    
            logger.info(f"Running EWS-Plotting with the following configs:\n{sys_config}\n{run_config}\n{chart_config}")
    
            env_suit_processor = EnvSuitPostProcessor()
            env_suit_processor.set_param_config_files(sys_params_file_arg = sys_config,
                                                      chart_params_file_arg = chart_config,
                                                      run_params_file_arg = run_config,
                                                      es_output_dir_arg = output_dir,
                                                      issue_date_arg = config['StartString'],
                                                      disease_nc_template_arg = rie_template_dir)
    
            env_suit_processor.run_params.FILTER_FOR_COUNTRY = (filter_for_country.upper() == "TRUE")
    
            # Include further diseases in plotting. In this case the irrigated suitabilite for the rusts.
            # TODO: move this part out into a config
            extra_diseases = [
                EnvSuitDiseaseInfo("Stem rust temp-only", "stem_rust_temponly", config['StartString'], "StemRust_TempOnly", rie_template_dir),
                EnvSuitDiseaseInfo("Leaf rust temp-only", "leaf_rust_temponly", config['StartString'], "LeafRust_TempOnly", rie_template_dir),
                EnvSuitDiseaseInfo("Stripe rust temp-only", "stripe_temponly", config['StartString'], "StripeRust_TempOnly", rie_template_dir),
            ]
    
            env_suit_processor.add_diseases(diseases=extra_diseases)
    
            env_suit_processor.process()
    
            # check the output
            ews_plotting_output_dir = f"{output_dir}/images/"
            #EWSPlottingOutputGlobs += [
            #        # daily plots
            #        f"{EWSPlottingOutputDir}Daily/suitability_{region.lower()}_*_rust_daily_20*.png",
            #        # weekly plots
            #        f"{EWSPlottingOutputDir}Weekly/suitability_{region.lower()}_*_rust_total_20*.png"]
    
            ews_plotting_output_globs = [f"{ews_plotting_output_dir}*"]
    
        # check the output
        ews_plotting_output_globs = get_only_existing_globs(ews_plotting_output_globs,inplace=False)
    
        # check there is some output from EWS-plotting
        if not ews_plotting_output_globs:
            logger.error('EWS-Plotting did not produce any output')
            raise RuntimeError
    
        # provide list for transfer
        ews_plotting_outputs: [str] = sorted([file for glob_str in ews_plotting_output_globs for file in glob(glob_str)])
    
        return ews_plotting_outputs


if __name__ == '__main__':
    processor = ProcessorEnvironment()
    processor.run_processor("Environment")
