#ProcessorUtils.py
'''Functions called by Processor.py and ProcessorComponents.py . '''

import datetime
import glob
import json
import logging
import logging.config
import os
import re
from pathlib import Path
import shutil
from string import Template
import subprocess
import sys
import tarfile
from typing import List
import netCDF4 as nc
import numpy as np
import pandas as pd
import cf_units

from flagdir import jobStatus # created by jws52

logger = logging.getLogger(__name__)

loglevels = {'debug':    logging.DEBUG,
             'info':     logging.INFO,
             'warning':  logging.WARNING,
             'error':    logging.ERROR,
             'critical': logging.CRITICAL}

short_name = {
        'Advisory' : 'ADVISORY',
        'Deposition' : 'DEPOSITION',
        'Environment' : 'ENVIRONMENT_2.0',
        'Epidemiology' : 'EPI',
        'Survey' : 'SURVEYDATA',
        'Scraper' : 'SCRAPER',
        'MetResample' : 'MET_RESAMPLE',
        }

disease_latin_name_dict = {
        'StemRust' : 'P_GRAMINIS',
        'StripeRust' : 'P_STRIIFORMIS',
        'LeafRust' : 'P_RECONDITA',
        'WheatBlast' : 'M_ORYZAE'}

# define logging filter to obscure ODK passwords

def filter_exception_message(exc_info):
    '''To mask out passwords from subprocess input list.'''

    error = exc_info[1]

    # subprocess input appears as a list
    if len(error.args) > 1:
        if isinstance(error.args[1],list):

            msg_list = error.args[1]

            pwd_arg = '--odk_password'
            
            if pwd_arg in msg_list:
                
                # get the index of the password, it comes after the input flag
                passwd_flag_idx = [i for i,val in enumerate(msg_list) if val==pwd_arg][0]
                passwd_idx = passwd_flag_idx + 1

                # replace the password
                msg_list[passwd_idx] = '***'
    
    return

class PasswordODKFilter(logging.Filter):
    '''Built with 
    https://docs.python.org/3.6/howto/logging-cookbook.html#configuring-filters-with-dictconfig
    '''

    def __init__(self, param=None):
        # example of how to pass parameters
        self.param = param
    
    def filter(self, record):

        # on the primary message
        input = record.msg
        masked_msg = re.sub("password', '\S+'","password', '***'",input)
        record.msg = masked_msg

        # on any exception info
        if record.exc_info is not None:

            filter_exception_message(record.exc_info)

        # return boolean to determine whether record gets logged
        return True


def add_filters_to_sublogger(logger_case):

    for h in logger_case.handlers:

        h.addFilter(PasswordODKFilter())


def open_and_check_config(configFile):
    logger.info("Loading config")

    try:
        with open(configFile) as config_file:
            config = json.load(config_file)
    except:
        logger.exception(f"Failure opening config file {configFile}")
        raise

    logger.debug(json.dumps(config,indent=2))

    #TODO: Checking config file for required formatting
    try:
        pass
    except AssertionError as e:
        logger.exception("Configuration file is not formatted as required")
        raise

    return config


def get_only_existing_globs(file_globs: [str], inplace = True):
    
    if inplace:
        for i,fg in enumerate(file_globs):
            if not glob.glob(fg):
                logger.warning(f'This file pattern returns nothing and therefore does not contribute to result:\n{fg}')
                file_globs.pop(i)
    else:
        globs_out = []
        for i,fg in enumerate(file_globs):
            if not glob.glob(fg):
                logger.warning(f'This file pattern returns nothing and therefore does not contribute to result:\n{fg}')
                continue
            globs_out += [fg]
        return globs_out

def subprocess_and_log(cmd,description_short,description_long,check=True,log_type='error', shell: bool = False, **kwargs):
    '''Run a shell command (described by a comma separated list) and send stdout
    and stderr to logfile, and raise any exception.
    
    log_type is a string to determine which log message level to use, must be 
    one of the logging options: info, debug, warn, warning, error, exception or
    fatal.'''

    # determine message type to write to log
    log_call = getattr(logger,log_type)

    try:
        process = subprocess.run(
                cmd,
                check=check,
                shell=shell,
                stdout = subprocess.PIPE,
                stderr = subprocess.STDOUT,
                **kwargs)

        for line in process.stdout.decode('utf-8').split(r'\n'):
            logger.info(f"{description_short} : " + line)

    except subprocess.CalledProcessError as e:

        for line in e.stdout.decode('utf-8').split(r'\n'):
            logger.info(f"{description_short} : " + line)

        log_call(f"Some failure when running {description_long}", exc_info=True)
        raise

    return process

def append_item_to_list(
        item: list,
        li: list,
        description: str,
        status: jobStatus) -> None:

    if isinstance(item,list):
        li += item
    elif item is None:
        pass
    else:
        logger.error(f"Misunderstood output of {description}()")
        status.reset('ERROR')
        endJob(status,premature=True)

    return
    
def clear_up(
        clearup_paths: List = None,
        clearup_dest: str = './clearup/') -> None:
    """ For now, moves the specified paths to the destination directory.
    TODO: Remove the specified paths.
    
    CAUTION: It is always preferable to avoid creating files in the first
    place."""

    assert isinstance(clearup_paths,list)

    logger.info(f"Clearing up the following files/directories: {clearup_paths}")

    Path(clearup_dest).mkdir(parents=True, exist_ok=True)

    for path in clearup_paths:
        basename = os.path.basename(path)
        clearup_dest_path = f"{clearup_dest}/{basename}"
        shutil.move(path,clearup_dest_path)

    return    

def endScript(
        premature: bool = True,
        ):

    if not premature:

        logger.info(f'Script finished!')
        exit_code = 0
    else:

        logger.info(f'Script finished prematurely')
        exit_code = 1

    logger.info(f'--------')
    sys.exit(exit_code)

def endJob(status,ignore_inprogress=False,**kwargs):

    # get status file fresh, and ensure one exists
    status.status = status.get()

    assert status.has_status

    if status.is_inprogress() & (ignore_inprogress == False):

        logger.warning(f'Workflow problem: status is {status.status} but endJob called, forcing error status')

        status.reset('ERROR')

    endScript(**kwargs)

def calc_epi_date_range(init_str, span_days = [0, 6]):
    '''Date range is  determined relative to init_date.
    span_days is usually defined in the job config file. Day zero is current
    day, negative values point to past (historical or analysis) days, and
    positive values point to forecast days.
    Returns a start_date and end_date.'''

    init_date = datetime.datetime.strptime(init_str,'%Y%m%d')

    # note that filename date represents preceding 3 hours, so day's data
    # starts at file timestamp 0300 UTC
    threehour_shift = datetime.timedelta(hours=3)

    # add 24hrs so that final day is fully included
    day_shift = datetime.timedelta(days=1)

    # if more than 999 days
    if len(str(span_days[0]))>3:
        # assume it is a date string
        start_date = datetime.datetime.strptime(span_days[0]+'0300','%Y%m%d%H%M')
    else:
        date_shift0 = datetime.timedelta(days=span_days[0])

        start_date = init_date + date_shift0 + threehour_shift

    if len(str(span_days[1]))>3:
        # assume it is a date string
        end_date = datetime.datetime.strptime(span_days[1]+'0000','%Y%m%d%H%M')

        end_date = end_date + day_shift
    else:
        date_shift1 = datetime.timedelta(days=span_days[1])

        end_date = init_date + date_shift1 +day_shift

    return start_date, end_date

def remove_path_from_tar_members(tf: tarfile.TarFile):
    """
    strips the parent path from files within a tar file before untaring. means untarred files will not have subdirs.
    Only used with the tarfile package
    """
    for member in tf.getmembers():
        member.path = os.path.basename(member.path)
        yield member

def query_proceed(necessary_file,description):

    try:

        assert os.path.isfile(necessary_file)

        logger.info(f"Found:\n{necessary_file}\nso {description} job has succeeded for this date, this job shall run.")

    except AssertionError as e:

        logger.info(f"Failed to find:\n{necessary_file}\nso {description} job has not yet succeeded for this date, so cannot run this job.")

        endScript(premature=True)

        return False

    return True

def query_component_success(config_i,job_run: str, job_to_check: str):
    '''Checks if a coordinator jobs has already completed successfully. If not,
    it raises an error.'''

    # check if deposition data is readily available
    status_file = Template(config_i[job_run][job_to_check]['SuccessFileTemplate']).substitute(**config_i)
    try:
        query_proceed(status_file,job_to_check)
    except:
        
        if 'AlternativeSuccessFileTemplate' not in config_i[job_run][job_to_check]:
            logger.warning(f"No AlternativeSuccessFileTemplate to check for")
            endScript(premature=True)

        status_file_alt = Template(config_i[job_run][job_to_check]['AlternativeSuccessFileTemplate']).substitute(**config_i)
        query_proceed(status_file_alt,job_to_check.lower())

    return True

def dataframe_to_series(df: pd.DataFrame) -> pd.Series:
    """Reformate a pandas Dataframe into a pandas Series.
    
    The dataframe should have the following structure:
    - first two columns: longitude and latitude
    - following columns: timestamps in YYYYMMDDhhmm format (e.g. 202311211500)
    
    The resulting Series will have the following structure:
    - indexes: time, longitude, latitude."""

    # longitude and latitude as index
    df=df.set_index(["longitude","latitude"])

    # convert dataframe to Series by stacking
    s=df.stack()

    # rename indexes
    s.index.names=["longitude","latitude","time"]

    # reorder index
    s=s.reorder_levels(["time","latitude","longitude"])

    # sort by timestamps
    s=s.sort_index(level=0)
    logger.info("Dataframe reformated to Series")

    return s

def create_bounds(coord):
    """Create bounds for a coordinate based on the difference between two adjacent values.
    Works for non equidistant coordinates as well."""

    # calculate the difference between two adjacent values and divide by 2
    diff = np.diff(coord)/2

    # insert the first and last value at the beginning and end of the array
    diff_lower = np.insert(diff, 0, diff[0])
    diff_upper = np.append(diff, diff[-1])

    # bounds are the coordinate values minus/plus the difference
    bounds = np.zeros((len(coord), 2))
    bounds[:, 0] = coord - diff_lower
    bounds[:, 1] = coord + diff_upper

    return bounds

def check_cf(units: str) -> str:
    """Check if units are CF convention compliant."""
    try: cf_units.Unit(units)
    except: raise ValueError(f"Units \"{units}\" not CF convention compliant!")
    return units

def create_nc_file(s: pd.Series, filename: str, data_attr: dict, compression_type="zlib", complevel=9, least_significant_digit=None):
    """Create a netcdf file from a pandas Series.

    The Series should have the following indexes:
    - time: str
    - latitude: float
    - longitude: float

    Path to the netcdf file should be filename: str

    Dictionary data_attr should have the following structure:
    { "short_name": str,
      "long_name": str
      "units": str,
      "description": str}

    Optional arguments:
    - compression_type: str, default="zlib"
    - least_significant_digit: int, default=5
    """

    ## Get coordinates #########################
    times=s.index.levels[0].values
    # convert times to hours since 1970-01-01 00:00:00
    times=pd.to_datetime(times).astype(np.int64)//((10**9)*60*60)

    latitudes=s.index.levels[1].unique().values
    longitudes=s.index.levels[2].unique().values

    ## Create bounds ###########################
    lat_bnds=create_bounds(latitudes)
    lon_bnds=create_bounds(longitudes)

    ## Create netcdf file ######################
    logger.info("Creating netcdf file")
    # just to be safe, make sure dataset is not already open.
    try: nc.Dataset(filename,"w",format="NETCDF4").close()  
    except: pass

    ncFile=nc.Dataset(filename,"w",format="NETCDF4")

    # create dimensions
    ncFile.createDimension("time", None)
    ncFile.createDimension("bnds",2)
    ncFile.createDimension("latitude",len(latitudes))
    ncFile.createDimension("longitude",len(longitudes))

    # create global attributes
    ncFile.description=data_attr["description"]
    ncFile.convension="CF-1.10"

    # create variables
    time=ncFile.createVariable("time",np.int32,("time",))
    time.axis="T"
    time.units="hours since 1970-01-01 00:00:00"
    time.long_name="time"
    time.calendar="gregorian"

    latitude_bnds=ncFile.createVariable("latitude_bnds",np.float32,("latitude","bnds",))
    latitude_bnds.long_name="latitude boundaries"
    latitude_bnds.units="degrees_north"

    latitude=ncFile.createVariable("latitude",np.float32,("latitude",))
    latitude.axis="Y"
    latitude.bounds="latitude_bnds"
    latitude.units="degrees_north"
    latitude.long_name="latitude"
    latitude.standard_name="grid_latitude"
    latitude.valid_min=latitudes.min().astype(np.float32)
    latitude.valid_max=latitudes.max().astype(np.float32)

    longitude_bnds=ncFile.createVariable("longitude_bnds",np.float32,("longitude","bnds",))
    longitude_bnds.long_name="longitude boundaries"
    longitude_bnds.units="degrees_east"

    longitude=ncFile.createVariable("longitude",np.float32,("longitude",))
    longitude.axis="X"
    longitude.bounds="longitude_bnds"
    longitude.units="degrees_east"
    longitude.long_name="longitude"
    longitude.standard_name="grid_longitude"
    longitude.valid_min=longitudes.min().astype(np.float32)
    longitude.valid_max=longitudes.max().astype(np.float32)

    latitude_longitude=ncFile.createVariable("latitude_longitude",np.int32)
    latitude_longitude.grid_mapping_name="latitude_longitude"
    latitude_longitude.longitude_of_prime_meridian=0.0
    latitude_longitude.earth_radius=6371229.0


    # create DATA variables
    logger.info("Creating data variable")
    DATA=ncFile.createVariable(data_attr["short_name"],np.float32,("time","latitude","longitude",), compression=compression_type, complevel=complevel, least_significant_digit=least_significant_digit)
    DATA.short_name=data_attr["short_name"]
    DATA.long_name=data_attr["long_name"]
    DATA.units=check_cf(data_attr["units"])
    DATA.description=data_attr["description"]
    DATA.grid_mapping="latitude_longitude"
    DATA.coordinates="time latitude longitude"

    # write data to variables
    time[:]=times
    latitude_bnds[:]=lat_bnds
    latitude[:]=latitudes
    longitude_bnds[:]=lon_bnds
    longitude[:]=longitudes

    # map data to DATA variable
    DATA[:]=s.values.reshape(len(times),len(latitudes),len(longitudes))

    ncFile.close()
    logger.info(f"{filename} netCDF4 file created.")
    return True

def setup_logging(job_file_path: str, is_live: bool, log_level: str):

    """
    sets up the Logging configuration for the project by loading the config json file (specified as an environment
    variable).
    :return:
    """

    assert 'LOGGING_CONFIG' in os.environ
    logging_config_env = os.environ['LOGGING_CONFIG']
    assert os.path.exists(logging_config_env)

    # get the email credentials file path from the environment variables
    assert 'EMAIL_CRED' in os.environ
    email_credential_fn = os.environ['EMAIL_CRED']
    assert os.path.exists(email_credential_fn)

    with open(email_credential_fn, 'r') as f:
        gmail_config = json.load(f)

    # check contents
    required_keys = ['user', 'pass', 'host', 'port', 'toaddrs']
    for required_key in required_keys:
        assert required_key in gmail_config

    arg_string = ' '.join(sys.argv)

    # load logging json config file
    json_file_path = logging_config_env
    with open(json_file_path, 'r') as f:
        log_config_dict = json.load(f)

    # replace the {arg_string} roken in the logging json with the values passed in the command line. This allows the
    # command line arguments to be logged.
    log_config_dict['formatters']['detailed']['format'] = \
        log_config_dict['formatters']['detailed']['format'].replace("{arg_string}", arg_string)

    log_config_dict['filters']['mask_passwords']['()'] = PasswordODKFilter

    log_config_dict['handlers']['handler_job']['filename'] = job_file_path

    # if there is no value set for the central project logger in the json file, set a default path.
    if not log_config_dict['handlers']['handler_project']['filename']:
        # todo how does this work when there are several processors running at once, i.o. errors?
        log_path_project = os.path.join(os.path.dirname(__file__), "logs", "log.txt")
        print(f"'HANDLER_PROJECT' HAS NO 'FILENAME' SET TO SETTING TO {log_path_project}")
        log_config_dict['handlers']['handler_project']['filename'] = log_path_project

    if is_live:
        # TODO: smtp handler can only use tls, but ssl is more secure. Look into defining/writing a suitable smtp handler
        log_config_dict['handlers']['handler_buffered_email']['server'] = (gmail_config['host'], gmail_config['port'])
        log_config_dict['handlers']['handler_buffered_email']['credentials'] = (gmail_config['user'], gmail_config['pass'])
        log_config_dict['handlers']['handler_buffered_email']['fromaddr'] = gmail_config['user']
        log_config_dict['handlers']['handler_buffered_email']['toaddrs'] = gmail_config['toaddrs']
    else:
        # if not live - we dont want to send emails
        print(f"is_live = False, so removing email handler from logging config")
        log_config_dict['handlers'].pop('handler_buffered_email')
        for logger_key in log_config_dict['loggers'].keys():
            handlers = log_config_dict['loggers'][logger_key]['handlers']
            if 'handler_buffered_email' in handlers:
                handlers.remove('handler_buffered_email')

    assert log_level in loglevels
    for logger_key in log_config_dict['loggers'].keys():
        log_config_dict['loggers'][logger_key]['level'] = loglevels.get(log_level)

    logging.config.dictConfig(log_config_dict)
    logger.info("Logging setup complete")
